from imblearn.combine import SMOTETomek
from imblearn.over_sampling import SMOTE
from imblearn.under_sampling import RandomUnderSampler
from sklearn.experimental import enable_iterative_imputer  # noqa
from sklearn.impute import IterativeImputer, KNNImputer, SimpleImputer
from sklearn.preprocessing import RobustScaler, StandardScaler

imputer_mapping = {
    "iterative": IterativeImputer,
    "knn": KNNImputer,
    "mean": lambda: SimpleImputer(strategy="mean"),
    "median": lambda: SimpleImputer(strategy="median"),
    "constant": lambda: SimpleImputer(strategy="constant"),
}

scaler_mapping = {
    "standard": StandardScaler,
    "robust": RobustScaler,
}

resampler_mapping = {
    "SMOTE": SMOTE,
    "undersample": RandomUnderSampler,
    "combine": SMOTETomek,
}



import logging
import os
import time
from copy import copy
from typing import List, Literal, Optional

import joblib
import numpy as np
import pandas as pd
from sklearn.base import clone
from sklearn.ensemble import (
    RandomForestClassifier,
    StackingClassifier,
    VotingClassifier,
)
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import LabelEncoder

from mamut.preprocessing.preprocessing import Preprocessor
from mamut.utils.utils import metric_dict

from .evaluation import ModelEvaluator
from .model_selection import ModelSelector

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class Mamut:
    """
    A class used to manage the machine learning pipeline, including preprocessing, model selection, and evaluation.

    Attributes
    ----------
    preprocess : bool
        Whether to apply preprocessing to the data.
    imb_threshold : float
        Threshold for detecting imbalanced data.
    exclude_models : Optional[List[str]]
        List of models to exclude from selection.
    score_metric : callable
        Metric used to evaluate model performance.
    optimization_method : Literal["random_search", "bayes"]
        Method for hyperparameter optimization.
    n_iterations : Optional[int]
        Number of iterations for optimization.
    random_state : Optional[int]
        Random state for reproducibility.
    preprocessor : Preprocessor
        Preprocessor object for data preprocessing.
    le : LabelEncoder
        Label encoder for target variable.
    model_selector : ModelSelector
        Object for model selection.
    X : pd.DataFrame
        Input features.
    y : pd.Series
        Target variable.
    X_train : pd.DataFrame
        Training features.
    X_test : pd.DataFrame
        Test features.
    y_train : pd.Series
        Training target variable.
    y_test : pd.Series
        Test target variable.
    raw_fitted_models_ : Optional[List[Pipeline]]
        List of raw fitted models.
    fitted_models_ : Optional[List[Pipeline]]
        List of fitted models with preprocessing.
    best_model_ : Optional[Pipeline]
        Best model pipeline.
    best_score_ : float
        Best model score.
    training_summary_ : dict
        Summary of the training process.
    optuna_studies_ : dict
        Optuna studies for hyperparameter optimization.
    ensemble_ : Optional[Pipeline]
        Ensemble model pipeline.
    greedy_ensemble_ : Optional[Pipeline]
        Greedy ensemble model pipeline.
    ensemble_models_ : Optional[List[Pipeline]]
        List of models in the ensemble.
    imbalanced_ : bool
        Whether the data is imbalanced.

    Methods
    -------
    fit(X: pd.DataFrame, y: pd.Series) -> Pipeline
        Fits the model to the data.
    predict(X: pd.DataFrame) -> np.ndarray
        Predicts the target variable for the given data.
    predict_proba(X: pd.DataFrame) -> np.ndarray
        Predicts the probabilities of the target variable for the given data.
    evaluate() -> None
        Evaluates the fitted models.
    save_best_model(path: str) -> None
        Saves the best model to the specified path.
    create_ensemble(voting: Literal["soft", "hard"] = "soft") -> Pipeline
        Creates an ensemble of the fitted models.
    create_greedy_ensemble(n_models: int = 6, voting: Literal["soft", "hard"] = "soft") -> Pipeline
        Creates a greedy ensemble of the fitted models.
    """

    def __init__(
        self,
        preprocess: bool = True,
        imb_threshold: float = 0.10,
        exclude_models: Optional[List[str]] = None,
        score_metric: Literal[
            "accuracy",
            "precision",
            "recall",
            "f1",
            "balanced_accuracy",
            "jaccard",
            "roc_auc_score",
        ] = "f1",
        optimization_method: Literal["random_search", "bayes"] = "bayes",
        n_iterations: Optional[int] = 30,
        random_state: Optional[int] = 42,
        **preprocessor_kwargs,
    ):
        """
        Constructs all the necessary attributes for the Mamut object.

        Parameters
        ----------
        preprocess : bool
            Whether to apply preprocessing to the data.
        imb_threshold : float
            Threshold for detecting imbalanced data.
        exclude_models : Optional[List[str]]
            List of models to exclude from selection.
        score_metric : Literal["accuracy", "precision", "recall", "f1", "balanced_accuracy", "jaccard", "roc_auc_score"]
            Metric used to evaluate model performance.
        optimization_method : Literal["random_search", "bayes"]
            Method for hyperparameter optimization.
        n_iterations : Optional[int]
            Number of iterations for optimization.
        random_state : Optional[int]
            Random state for reproducibility.
        **preprocessor_kwargs
            Additional keyword arguments for the Preprocessor.
        """
        self.preprocess = preprocess
        self.imb_threshold = imb_threshold
        self.exclude_models = exclude_models
        self.score_metric = metric_dict[score_metric]
        self.optimization_method = optimization_method
        self.n_iterations = n_iterations
        self.random_state = random_state

        self.preprocessor = Preprocessor(**preprocessor_kwargs) if preprocess else None
        self.le = LabelEncoder()
        self.model_selector = None

        self.X = None
        self.y = None
        self.X_train = None
        self.X_test = None
        self.y_train = None
        self.y_test = None
        self.binary = None
        self.roc = None

        self.raw_fitted_models_ = None
        self.fitted_models_ = None
        self.best_model_ = None

        self.best_score_ = None
        self.training_summary_ = None
        self.optuna_studies_ = None

        self.ensemble_ = None
        self.greedy_ensemble_ = None
        self.ensemble_models_ = None
        self.imbalanced_ = None

    def fit(self, X: pd.DataFrame, y: pd.Series):
        """
        Fits the model to the data.

        Parameters
        ----------
        X : pd.DataFrame
            The input features.
        y : pd.Series
            The target variable.

        Returns
        -------
        Pipeline
            The best model pipeline.
        """
        Mamut._check_categorical(y)
        if y.value_counts(normalize=True).min() < self.imb_threshold:
            self.imbalanced_ = True

        y = self.le.fit_transform(y)
        y = pd.Series(y)

        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.2, stratify=y, random_state=self.random_state
        )

        if self.preprocess:
            X_train, y_train = self.preprocessor.fit_transform(X_train, y_train)
            X_test = self.preprocessor.transform(X_test)

        self.X_train = X_train  # np.ndarray
        self.X_test = X_test  # np.ndarray
        self.y_train = y_train  # np.ndarray
        self.y_test = y_test  # pd.Series
        self.X = X
        self.y = y

        self.model_selector = ModelSelector(
            X_train,
            y_train,
            X_test,
            y_test,
            exclude_models=self.exclude_models,
            score_metric=self.score_metric,
            optimization_method=self.optimization_method,
            n_iterations=self.n_iterations,
            random_state=self.random_state,
        )

        (
            best_model,
            params_for_best_model,
            score_for_best_model,
            fitted_models,
            training_summary,
            studies,
        ) = self.model_selector.compare_models()

        self.raw_fitted_models_ = fitted_models
        self.optuna_studies_ = studies
        self.fitted_models_ = [
            Pipeline([("preprocessor", self.preprocessor), ("model", model)])
            for model in fitted_models.values()
        ]

        # Update the score metric based on binary/multiclass problem (for ensembles)
        self.score_metric = self.model_selector.score_metric
        self.score_metric_name = self.model_selector.score_metric_name
        self.binary = self.model_selector.binary
        self.roc = self.model_selector.roc
        self.best_score_ = score_for_best_model
        self.best_model_ = Pipeline(
            [("preprocessor", self.preprocessor), ("model", best_model)]
        )
        self.training_summary_ = training_summary

        log.info(f"Best model: {best_model.__class__.__name__}")

        # Models_dir with time signature
        cwd = os.getcwd()
        models_dir = os.path.join(
            cwd,
            "fitted_models",
            str(time.strftime("%Y-%m-%d_%H:%M:%S", time.localtime())),
        )
        os.makedirs(models_dir, exist_ok=True)
        for model in self.fitted_models_:
            model_name = model.named_steps["model"].__class__.__name__
            model_path = os.path.join(models_dir, f"{model_name}.joblib")
            joblib.dump(model, model_path)
            log.info(f"Saved model {model_name} to {model_path}")

        return self.best_model_

    def predict(self, X: pd.DataFrame):
        """
        Predicts the target variable for the given data.

        Parameters
        ----------
        X : pd.DataFrame
            The input features.

        Returns
        -------
        np.ndarray
            Predicted target variable.
        """
        return self._predict(X)

    def predict_proba(self, X: pd.DataFrame):
        """
        Predicts the probabilities of the target variable for the given data.

        Parameters
        ----------
        X : pd.DataFrame
            The input features.

        Returns
        -------
        np.ndarray
            Predicted probabilities of the target variable.
        """
        return self._predict(X, proba=True)

    def evaluate(self, n_top_models: int = 3) -> None:
        """
        Evaluates the fitted models.
        """
        self._check_fitted()

        evaluator = ModelEvaluator(
            self.raw_fitted_models_,
            X_test=self.X_test,
            y_test=self.y_test,
            X_train=self.X_train,
            y_train=self.y_train,
            X=self.X,
            y=self.y,
            optimizer=self.optimization_method,
            metric=self.score_metric_name,
            n_trials=self.n_iterations,
            excluded_models=self.exclude_models,
            studies=self.optuna_studies_,
            training_summary=self.training_summary_,
            pca_loadings=self.preprocessor.pca_loadings_,
            binary=self.model_selector.binary,
            preprocessing_steps=self.preprocessor.report(),
            n_top_models=n_top_models,
            is_ensemble=self.greedy_ensemble_ is not None,
            greedy_ensemble=self.greedy_ensemble_,
        )

        evaluator.evaluate_to_html(self.training_summary_)
        evaluator.plot_results_in_notebook()

    def save_best_model(self, path: str) -> None:
        """
        Saves the best model to the specified path.

        Parameters
        ----------
        path : str
            The path to save the best model.
        """
        self._check_fitted()
        save_path = os.path.join(
            path, f"{self.best_model_.named_steps['model'].__class__.__name__}.joblib"
        )
        joblib.dump(self.best_model_, save_path)
        log.info(f"Saved best model to {save_path}")

    def create_ensemble(self, voting: Literal["soft", "hard"] = "soft") -> Pipeline:
        """
        Creates an ensemble of the fitted models.

        Parameters
        ----------
        voting : Literal["soft", "hard"]
            Voting strategy for the ensemble.

        Returns
        -------
        Pipeline
            The ensemble model pipeline.
        """
        self._check_fitted()

        ensemble = VotingClassifier(
            estimators=[
                (
                    model.named_steps["model"].__class__.__name__,
                    clone(model.named_steps["model"]),
                )
                for model in self.fitted_models_
            ],
            voting=voting,
        )
        ensemble.fit(self.X_train, self.y_train)
        y_pred = ensemble.predict(self.X_test)
        score = self.score_metric(self.y_test.values, y_pred)

        self.ensemble_ = Pipeline(
            [("preprocessor", self.preprocessor), ("model", ensemble)]
        )
        log.info(
            f"Created ensemble with all models and voting='{voting}'. "
            f"Ensemble score on test set: {score:.4f} {self.score_metric.__name__}"
        )

        return self.ensemble_

    def _create_greedy_ensemble_voting(
        self, n_models: int = 6, voting: Literal["soft", "hard"] = "soft"
    ) -> Pipeline:
        """
        Creates a greedy ensemble of the fitted models.

        Parameters
        ----------
        n_models : int
            Number of models to include in the ensemble.
        voting : Literal["soft", "hard"]
            Voting strategy for the ensemble.

        Returns
        -------
        Pipeline
            The greedy ensemble model pipeline.
        """
        self._check_fitted()

        # Initialize the ensemble with the best model
        ensemble_models = [self.best_model_.named_steps["model"]]
        ensemble_scores = [self.best_score_]

        for _ in range(n_models - 1):
            best_score = 0
            best_model = None

            for model in self.fitted_models_:
                candidate_ensemble = ensemble_models + [model.named_steps["model"]]
                candidate_voting_clf = VotingClassifier(
                    estimators=[
                        (f"model_{i}", clone(m))
                        for i, m in enumerate(candidate_ensemble)
                    ],
                    voting=voting,
                )
                candidate_voting_clf.fit(self.X_train, self.y_train)
                score = self.score_metric(
                    self.y_test, candidate_voting_clf.predict(self.X_test)
                )

                if score > best_score:
                    best_score = score
                    best_model = model.named_steps["model"]

            ensemble_models.append(best_model)
            ensemble_scores.append(best_score)

        ensemble = VotingClassifier(
            estimators=[
                (f"model_{i}", clone(m)) for i, m in enumerate(ensemble_models)
            ],
            voting=voting,
        )
        ensemble.fit(self.X_train, self.y_train)
        y_pred = ensemble.predict(self.X_test)
        score = self.score_metric(self.y_test, y_pred)

        self.ensemble_models_ = ensemble_models
        self.greedy_ensemble_ = Pipeline(
            [("preprocessor", self.preprocessor), ("model", ensemble)]
        )

        log.info(
            f"Created greedy ensemble with voting='{voting}' \n"
            f"and {n_models} models: {[m.__class__.__name__ for m in ensemble_models]} \n"
            f"Ensemble score on test set: {score:.4f} {self.score_metric.__name__}"
        )

        return self.greedy_ensemble_

    def create_greedy_ensemble(self, max_models=6):
        self._check_fitted()
        models = [model for name, model in self.raw_fitted_models_.items()]

        if max_models > len(models):
            max_models = len(models)
            log.info(
                f"Max models set to {max_models} as there are only {len(models)} models available"
                f"in the bag-of-models used in this experiment."
            )

        # Sort models list by their performance on X_test
        sorted_models = sorted(
            models, key=lambda model: self._score_model_on_test(model), reverse=True
        )

        # Start with the best and second best model
        ensemble_models = [sorted_models[0], sorted_models[1]]
        remaining_models = {
            model.__class__.__name__: model for model in copy(sorted_models[2:])
        }
        best_score = self._score_model_on_test(
            self._create_stacking_classifier(ensemble_models).fit(
                self.X_train, self.y_train
            )
        )
        ensemble_scores = [best_score]

        # Greedily add models to the ensemble
        for _ in range(max_models - 2):
            best_score = 0
            best_model = None

            for model in remaining_models.values():
                candidate_ensemble = ensemble_models + [model]
                candidate_stacking_clf = self._create_stacking_classifier(
                    candidate_ensemble
                )
                candidate_stacking_clf.fit(self.X_train, self.y_train)
                score = self._score_model_on_test(candidate_stacking_clf)

                if score > best_score:
                    best_score = score
                    best_model = model

            ensemble_models.append(best_model)
            ensemble_scores.append(best_score)
            # Remove this best model from the remaining models dict
            del remaining_models[best_model.__class__.__name__]

        # From nested family of ensembles pick the best one based on ensemble_scores
        best_score = max(ensemble_scores)
        # Check from the end of the list to find the best ensemble
        for i in range(len(ensemble_scores) - 1, 0, -1):
            if ensemble_scores[i] == best_score:
                ensemble_models = ensemble_models[: i + 2]
                break

        # Create the final stacking classifier
        final_stacking_clf = self._create_stacking_classifier(ensemble_models)
        final_stacking_clf.fit(self.X_train, self.y_train)
        self.greedy_ensemble_ = final_stacking_clf

        log.info(
            f"Created greedy ensemble with {len(ensemble_models)} models. Best score: {best_score:.4f}."
            f"For details on the ensemble please run evaluate() method and see the report."
        )

        # Create a pipeline with the best ensemble
        self.greedy_ensemble_ = Pipeline(
            [("preprocessor", self.preprocessor), ("model", final_stacking_clf)]
        )

        return self.greedy_ensemble_

    def _score_model_on_test(self, model):
        if self.roc:
            if self.binary:
                score_on_test = self.score_metric(
                    self.y_test.values, model.predict_proba(self.X_test)[:, 1]
                )
            else:
                score_on_test = self.score_metric(
                    self.y_test.values, model.predict_proba(self.X_test)
                )
        else:
            score_on_test = self.score_metric(
                self.y_test.values, model.predict(self.X_test)
            )
        return score_on_test

    def _create_stacking_classifier(self, models):
        estimators = [(model.__class__.__name__, clone(model)) for model in models]
        return StackingClassifier(
            estimators=estimators, final_estimator=RandomForestClassifier()
        )

    def _calculate_disagreement(self, model1, model2, X_test):
        #  TODO: THIS IS WORK IN PROGRESS... DO NOT USE
        """Calculate disagreement between two models' predictions."""
        pred1 = model1.predict(X_test)
        pred2 = model2.predict(X_test)
        return np.mean(pred1 != pred2)

    def _ensemble_selection(
        self, max_ensemble_size=5, voting: Literal["soft", "hard"] = "soft"
    ):
        """Greedy algorithm to select the best subset of models for ensemble."""
        #  TODO: THIS IS WORK IN PROGRESS... DO NOT USE
        models = [
            (model.__class__.__name__, model)
            for model in self.raw_fitted_models_.values()
        ]
        print(models)
        selected_models = []
        remaining_models = models.copy()
        best_score = 0
        ensemble_performance = []

        # Initialize with the best performing model on test set
        scores = {
            name: self.score_metric(self.y_test, model.predict(self.X_test))
            for name, model in models
        }
        print("Scores:", scores)
        best_model_name, best_model = max(scores.items(), key=lambda item: item[1])
        print("Best model:", best_model_name, "with score:", best_model)
        selected_models.append((best_model_name, best_model))
        remaining_models.remove((best_model_name, best_model))
        best_score = scores[best_model_name]
        ensemble_performance.append(best_score)

        print(f"Starting with best model: {best_model_name} with score: {best_score}")

        # Greedily add models based on performance and diversity
        while len(selected_models) < max_ensemble_size and remaining_models:
            best_model_to_add = None
            best_new_score = best_score
            for name, model in remaining_models:
                # Test current ensemble with this model added
                current_ensemble = VotingClassifier(
                    estimators=selected_models + [(name, model)], voting=voting
                )
                current_ensemble.fit(self.X_train, self.y_train)
                ensemble_score = self.score_metric(
                    self.y_test, current_ensemble.predict(self.X_test)
                )

                # Compute diversity with selected models
                diversity = np.mean(
                    [
                        self._calculate_disagreement(
                            model, selected_model[1], self.X_test
                        )
                        for selected_model in selected_models
                    ]
                )

                # Score considering both accuracy improvement and diversity
                weighted_score = (
                    ensemble_score + 0.1 * diversity
                )  # 0.1 is a diversity weight factor
                if weighted_score > best_new_score:
                    best_model_to_add = (name, model)
                    best_new_score = weighted_score

            if best_model_to_add:
                selected_models.append(best_model_to_add)
                remaining_models.remove(best_model_to_add)
                best_score = best_new_score
                ensemble_performance.append(best_new_score)
                print(
                    f"Added {best_model_to_add[0]} to ensemble, new weighted score: {best_new_score}"
                )
            else:
                break  # No improvement

        # Final ensemble
        final_ensemble = VotingClassifier(estimators=selected_models, voting=voting)
        final_ensemble.fit(self.X_train, self.y_train)
        return final_ensemble, ensemble_performance

    def _predict(self, X: pd.DataFrame, proba: bool = False):
        """
        Predicts the target variable or probabilities for the given data.

        Parameters
        ----------
        X : pd.DataFrame
            The input features.
        proba : bool
            Whether to predict probabilities instead of the target variable.

        Returns
        -------
        np.ndarray
            Predicted target variable or probabilities.
        """
        self._check_fitted()
        if proba:
            return self.best_model_.predict_proba(X)
        return self.best_model_.predict(X)

    def _check_fitted(self):
        """
        Checks if the model has been fitted.

        Raises
        ------
        RuntimeError
            If the model has not been fitted.
        """
        if not self.best_model_:
            raise RuntimeError(
                "Can't predict because no model has been fitted. "
                "Please call fit() method first."
            )

    @staticmethod
    def _check_categorical(y):
        """
        Checks if the target variable is categorical.

        Parameters
        ----------
        y : pd.Series
            The target variable.

        Raises
        ------
        ValueError
            If the target variable is not categorical.
        """
        if pd.api.types.is_float_dtype(y):
            raise ValueError("Target variable must be categorical.")



import base64
import math
import os
import platform
import time
import warnings
from datetime import datetime
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import optuna
import pandas as pd
import psutil
import seaborn as sns

warnings.filterwarnings(
    "ignore",
    category=DeprecationWarning,
    module="shap\\.plots\\.colors\\._colorconv",
)

import shap
from jinja2 import Environment, FileSystemLoader
from matplotlib import gridspec
from sklearn.ensemble import ExtraTreesClassifier
from sklearn.metrics import (
    accuracy_score,
    auc,
    balanced_accuracy_score,
    confusion_matrix,
    f1_score,
    jaccard_score,
    precision_score,
    recall_score,
    roc_auc_score,
    roc_curve,
)
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import label_binarize

from mamut.preprocessing.handlers import handle_outliers
from mamut.utils.utils import model_param_dict, preprocessing_steps


def _get_base64_image(image_path):
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")


def _generate_experiment_setup_table():
    system_info = {
        "Machine": platform.machine(),
        "Processor": platform.processor(),
        "System": platform.system(),
        "Release": platform.release(),
        "Version": platform.version(),
        "Python Version": platform.python_version(),
        "RAM (GB)": round(psutil.virtual_memory().total / (1024**3), 2),
        "CPU Cores": psutil.cpu_count(logical=True),
    }

    df = pd.DataFrame(system_info.items(), columns=["Attribute", "Value"])
    html_table = df.to_html(index=False)
    return html_table


def _generate_dataset_overview(
    X: pd.DataFrame, y: pd.Series
) -> (List[int], pd.DataFrame, pd.DataFrame):
    n_observations, n_features = X.shape

    n_rows_missing = X.isnull().any(axis=1).sum()

    # Calculate the number of outliers according to IsolationForest method:
    # TODO: Check if correct
    _, y_new, _ = handle_outliers(X, y, X.columns)
    n_outliers = len(y) - len(y_new)

    dataset_basic_list = [n_observations, n_features, n_rows_missing, n_outliers]

    feature_summary = X.dtypes.reset_index()
    feature_summary.columns = ["Feature", "Data Type"]
    feature_summary["Type"] = feature_summary["Data Type"].apply(
        lambda dt: (
            "Categorical"
            if dt == "object" or isinstance(dt, pd.CategoricalDtype)
            else "Numerical"
        )
    )
    if len(feature_summary) > 10:
        feature_summary = feature_summary.head(10)
    feature_summary = feature_summary[["Feature", "Type", "Data Type"]]

    if not isinstance(y, pd.Series):
        y = pd.Series(y)
    class_distribution = y.value_counts().reset_index()
    class_distribution.columns = ["Class", "Count"]

    return dataset_basic_list, feature_summary, class_distribution


def _generate_preprocessing_steps_list(steps) -> str:
    categorized_steps = {}

    for step in steps:
        if step in preprocessing_steps:
            category, description = preprocessing_steps[step]
            if category not in categorized_steps:
                categorized_steps[category] = []
            categorized_steps[category].append(
                f"<strong>{step}</strong>: {description}"
            )

    html_prep_list = ""
    for category, tools in categorized_steps.items():
        html_prep_list += f"<li style='padding-left: 10px;'><strong>{category}</strong><ul style='list-style-type: '&#x2192'; margin-left: 20px;'>"
        for tool in tools:
            html_prep_list += f"<li>{tool}</li>"
        html_prep_list += "</ul></li>"

    return html_prep_list


def _generate_preprocessing_steps_html(report):
    """
    Generates an HTML list of preprocessing steps based on the report dictionary.

    Parameters
    ----------
    report : dict
        The report dictionary containing preprocessing steps.

    Returns
    -------
    str
        An HTML string representing the preprocessing steps.
    """
    html_list = ""

    for category, steps in report.items():
        html_list += f"<li><strong>{category.capitalize()}</strong><ul>"
        if isinstance(steps, dict):
            for step, details in steps.items():
                if isinstance(details, dict):
                    html_list += f"<li><strong>{step.capitalize()}</strong><ul>"
                    for sub_step, description in details.items():
                        html_list += f"<li>{sub_step}: {description}</li>"
                    html_list += "</ul></li>"
                else:
                    html_list += f"<li>{step}: {details}</li>"
        else:
            html_list += f"<li>{steps}</li>"
        html_list += "</ul></li>"

    return html_list


def _generate_models_list(excluded_models: List[str]) -> List[str]:
    all_models = model_param_dict.keys()
    available_models = [model for model in all_models if model not in excluded_models]

    return available_models


def _generate_ensemble_list(ensemble: Pipeline) -> str:
    if not ensemble:
        return ""
    ensemble = ensemble.named_steps["model"]
    base_estimators = ensemble.estimators
    meta = ensemble.final_estimator
    # Generate HTML list with ensemble contents:
    html_list = ""

    html_list += "<li><strong>Base Estimators:</strong><ul>"
    for name, estimator in base_estimators:
        html_list += f"<li>{name}: {estimator.__class__.__name__}</li>"
    html_list += "</ul></li>"

    html_list += f"<li><strong>Meta Model:</strong> <ul><li>{meta.__class__.__name__}</li></ul></li>"

    return html_list


class ModelEvaluator:

    report_template_path: str = os.path.join(os.path.dirname(__file__), "utils")

    def __init__(
        self,
        models: dict,
        # X_test and y_test are preprocessed. X and y are not.
        X_test: np.ndarray,
        y_test: np.ndarray,
        X_train: np.ndarray,
        y_train: np.ndarray,
        X: pd.DataFrame,
        y: pd.Series,
        optimizer: str,
        n_trials: int,
        metric: str,
        studies: dict,
        training_summary: pd.DataFrame,
        pca_loadings,
        binary: bool,
        preprocessing_steps,
        is_ensemble: bool,
        greedy_ensemble,
        excluded_models: List[str] = None,
        n_top_models: int = 3,
    ):

        self.models = models
        self.X = X
        self.y = y
        self.X_test = X_test
        self.y_test = y_test
        self.X_train = X_train
        self.y_train = y_train
        self.optimizer = optimizer
        self.n_trials = n_trials
        self.metric = metric
        self.studies = studies
        self.training_summary = training_summary
        self.pca_loadings = pca_loadings
        self.binary = binary
        self.is_ensemble = is_ensemble
        self.greedy_ensemble = greedy_ensemble
        if self.pca_loadings is not None:
            self.pca = True
        else:
            self.pca = False
        if self.training_summary is None:
            raise ValueError(
                "You need to .fit() your models before evaluating them with .evaluate()"
            )
        self.preprocessing_steps = preprocessing_steps
        self.excluded_models = excluded_models if excluded_models else []

        self.report_output_path = os.path.join(os.getcwd(), "mamut_report")
        self.plot_output_path = os.path.join(self.report_output_path, "plots")

        self.n_top_models = n_top_models

        # Create the report directory it doesn't exist:
        os.makedirs(self.report_output_path, exist_ok=True)
        os.makedirs(self.plot_output_path, exist_ok=True)
        self._set_plt_style()

    def _set_plt_style(self) -> None:
        sns.set_context("notebook", font_scale=1.05)
        plt.style.use("fivethirtyeight")
        # Set background color of all plots to #f0f8ff;
        plt.rcParams["axes.facecolor"] = "#f0f8ff"
        plt.rcParams["figure.facecolor"] = "#f0f8ff"

    def evaluate(self, training_summary: pd.DataFrame):
        return self.evaluate_to_html(training_summary)

    def plot_results_in_notebook(self):
        if self.binary:
            self._plot_roc_auc_curve(
                training_summary=self.training_summary,
                n_top=self.n_top_models,
                show=True,
                save=False,
            )
        else:
            self._plot_roc_auc_curve_multiclass(
                training_summary=self.training_summary,
                n_top=self.n_top_models,
                show=True,
                save=False,
            )
        self._plot_confusion_matrices(
            n_top=self.n_top_models,
            show=True,
            save=False,
            training_summary=self.training_summary,
        )
        self._plot_hyperparameter_tuning_history(
            n_top=self.n_top_models,
            show=True,
            save=False,
            training_summary=self.training_summary,
        )
        return

    def _plot_roc_auc_curve(
        self,
        training_summary: pd.DataFrame,
        n_top: int = 3,
        show: bool = False,
        save: bool = True,
    ) -> None:
        fig, ax = plt.subplots(figsize=(12, 6))
        top_models = training_summary["Model"].head(n_top).to_numpy()

        for model_name in top_models:
            model = next(
                m for m in self.models.values() if m.__class__.__name__ == model_name
            )
            y_pred = model.predict_proba(self.X_test)[:, 1]
            fpr, tpr, thresholds = roc_curve(self.y_test, y_pred)
            auc = roc_auc_score(self.y_test, y_pred)
            ax.plot(fpr, tpr, lw=1.5, label=f"{model_name} ROC ({auc:.2f})")

        ax.plot([0, 1], [0, 1], "k--", lw=1.5)
        ax.set_xlim([-0.01, 1.01])
        ax.set_ylim([-0.01, 1.05])
        ax.set_xlabel("False Positive Rate", fontsize=12)
        ax.set_ylabel("True Positive Rate", fontsize=12)
        ax.legend(loc="lower right", fontsize=10)
        plt.tight_layout()

        if save:
            plt.savefig(
                os.path.join(self.plot_output_path, "roc_auc_curve.png"),
                format="png",
                bbox_inches="tight",
            )
        if show:
            plt.show()

        plt.close(fig)

        return

    def _plot_roc_auc_curve_multiclass(
        self,
        training_summary: pd.DataFrame,
        n_top: int = 3,
        show: bool = False,
        save: bool = True,
    ) -> None:
        fig, ax = plt.subplots(figsize=(12, 6))
        top_models = training_summary["Model"].head(n_top).to_numpy()
        y_test_bin = label_binarize(self.y_test, classes=np.unique(self.y_test))

        for model_name in top_models:
            model = next(
                m for m in self.models.values() if m.__class__.__name__ == model_name
            )
            y_score = model.fit(self.X_train, self.y_train).predict_proba(self.X_test)
            fpr, tpr, _ = roc_curve(y_test_bin.ravel(), y_score.ravel())
            roc_auc = auc(fpr, tpr)

            ax.plot(
                fpr,
                tpr,
                lw=2,
                label=f"Micro-averaged {model_name} (area = {roc_auc:0.2f})",
            )

        ax.plot([0, 1], [0, 1], "k--", lw=2)
        ax.set_xlim([-0.01, 1.0])
        ax.set_ylim([0.0, 1.05])
        ax.set_xlabel("False Positive Rate", fontsize=14)
        ax.set_ylabel("True Positive Rate", fontsize=14)
        ax.set_title("Micro-Averaged ROC Curve (One-vs-Rest)", fontsize=14)
        ax.legend(loc="lower right", fontsize=12)
        plt.tight_layout()

        if save:
            plt.savefig(
                os.path.join(self.plot_output_path, "roc_auc_curve.png"),
                format="png",
                bbox_inches="tight",
            )
        if show:
            plt.show()
        plt.close(fig)

        return

    def _plot_confusion_matrices(
        self,
        training_summary: pd.DataFrame,
        n_top: int = 3,
        show: bool = False,
        save: bool = True,
    ) -> None:
        rows = math.ceil(n_top / 3)
        fig = plt.figure(figsize=(18, 5 * rows))
        top_models = training_summary["Model"].head(n_top).to_numpy()
        if n_top == 3:
            gs = gridspec.GridSpec(1, 3, wspace=0.4)
        elif n_top > 3:
            gs = gridspec.GridSpec(rows, 3, wspace=0.3, hspace=0.3)
        else:
            gs = gridspec.GridSpec(1, n_top, wspace=0.3, hspace=0.3)

        for i, model_name in enumerate(top_models):
            model = next(
                m for m in self.models.values() if m.__class__.__name__ == model_name
            )
            y_pred = model.predict(self.X_test)
            cm = confusion_matrix(self.y_test, y_pred)

            ax = fig.add_subplot(gs[i])
            sns.heatmap(cm, annot=True, fmt="d", cmap="Blues", cbar=False, ax=ax)
            plt.title(f"{model_name}", fontsize=14)
            plt.xlabel("Predicted", fontsize=12)
            plt.ylabel("Actual", fontsize=12)

        plt.tight_layout()

        if save:
            plt.savefig(
                os.path.join(self.plot_output_path, "confusion_matrices.png"),
                format="png",
                bbox_inches="tight",
            )
        if show:
            plt.show()

        plt.close()

        return

    def _plot_hyperparameter_tuning_history(
        self,
        training_summary: pd.DataFrame,
        n_top: int = 3,
        show: bool = False,
        save: bool = True,
    ) -> None:
        self._set_plt_style()
        top_models = training_summary["Model"].head(n_top).to_numpy()

        for i, model_name in enumerate(top_models):
            study = self.studies.get(model_name)
            if study:
                plt.figure(figsize=(6, 5), facecolor="#f0f8ff")
                ax = optuna.visualization.matplotlib.plot_optimization_history(study)
                if not show:
                    ax.set_facecolor("#f0f8ff")
                    ax.spines["top"].set_color("#007bb5")
                    ax.spines["right"].set_color("#007bb5")
                    ax.spines["bottom"].set_color("#007bb5")
                    ax.spines["left"].set_color("#007bb5")
                    ax.grid(color="grey")  # Change grid color to grey
                ax.legend().set_visible(False)  # Remove legend
                plt.title(f"{model_name} Tuning History", fontsize=14)
                plt.xlabel("Trial", fontsize=12)
                plt.ylabel(f"{self.metric} Value", fontsize=12)
                plt.tight_layout()

                if save:
                    plt.savefig(
                        os.path.join(
                            self.plot_output_path,
                            f"hyperparameter_tuning_history_{i + 1}.png",
                        ),
                        format="png",
                        bbox_inches="tight",
                    )
                if show:
                    plt.show()

                plt.close()

        return

    def _plot_feature_importances(self, show: bool = False, save: bool = True) -> None:
        self._set_plt_style()
        rf = ExtraTreesClassifier(random_state=42)
        rf.fit(self.X_train, self.y_train)

        importances = rf.feature_importances_
        indices = np.argsort(importances)[::-1]

        if len(indices) > 10:
            indices = indices[:10]

        plt.figure(figsize=(10, 6))
        plt.bar(range(len(indices)), importances[indices], align="center")
        plt.xticks(range(len(indices)), self.X.columns[indices], rotation=90)
        plt.xlabel("Feature", fontsize=12)
        plt.ylabel("Importance", fontsize=12)
        plt.tight_layout()

        if save:
            plt.savefig(
                os.path.join(self.plot_output_path, "feature_importance.png"),
                format="png",
                bbox_inches="tight",
            )
        if show:
            plt.show()

        plt.close()

        return

    def _plot_shap_beeswarm(self, model, show: bool = False, save: bool = True) -> None:
        if model.__class__.__name__ in ["KNeighborsClassifier", "SVC", "MLPClassifier"]:
            explainer = shap.Explainer(model.predict, self.X_train)
        else:
            explainer = shap.Explainer(model, self.X_train)
        shap_values = explainer(self.X_train)

        if len(shap_values.shape) == 3:
            num_classes = shap_values.shape[2]
            for class_idx in range(num_classes):
                plt.figure(figsize=(10, 6))
                shap.plots.beeswarm(shap_values[:, :, class_idx], show=False)
                plt.title(f"SHAP Beeswarm Plot For Class {class_idx}", fontsize=14)
                plt.tight_layout()

                if save:
                    plt.savefig(
                        os.path.join(
                            self.plot_output_path,
                            f"shap_beeswarm_class_{class_idx}.png",
                        ),
                        format="png",
                        bbox_inches="tight",
                    )
                if show:
                    plt.show()

                plt.close()
        else:
            plt.figure(figsize=(10, 6))
            shap.plots.beeswarm(shap_values, max_display=10, show=False)
            plt.title("SHAP Beeswarm Plot", fontsize=14)
            plt.tight_layout()

            if save:
                plt.savefig(
                    os.path.join(self.plot_output_path, "shap_values.png"),
                    format="png",
                    bbox_inches="tight",
                )
            if show:
                plt.show()
            plt.close()
        return

    def _plot_pca_loadings(self, show: bool = False, save: bool = True) -> None:
        if self.pca_loadings is None:
            raise ValueError(
                "PCA loadings are not available. "
                "Potentially PCA was not used in the preprocessing steps."
                "Use Mamut(pca=True) to include PCA in the preprocessing steps."
            )

        self._set_plt_style()
        sns.set_palette(sns.color_palette("tab20", 20))

        plt.figure(figsize=(12, 8))
        sns.heatmap(
            self.pca_loadings,
            annot=False,
            cmap="coolwarm",
            xticklabels=self.X.columns,
            yticklabels=[f"PC{i + 1}" for i in range(self.pca_loadings.shape[0])],
        )
        plt.xlabel("Features", fontsize=12)
        plt.ylabel("Principal Components", fontsize=12)
        plt.tight_layout()

        if save:
            plt.savefig(
                os.path.join(self.plot_output_path, "pca_loadings_heatmap.png"),
                format="png",
                bbox_inches="tight",
            )
        if show:
            plt.show()
        plt.close()

        return

    def _plot_pca_loadings2(self, show: bool = False, save: bool = True) -> None:
        if self.pca_loadings is None:
            raise ValueError(
                "PCA loadings are not available. "
                "Potentially PCA was not used in the preprocessing steps."
                "Use Mamut(pca=True) to include PCA in the preprocessing steps."
            )
        self._set_plt_style()
        sns.set_palette(sns.color_palette("tab20", 20))
        n_components = self.pca_loadings.shape[0]
        n_features = self.pca_loadings.shape[1]

        plt.figure(figsize=(10, 6))
        for i in range(n_components):
            plt.bar(
                np.arange(n_features) + i / n_components,
                self.pca_loadings[i],
                width=1 / n_components,
                label=f"PC{i + 1}",
            )

        plt.xlabel("Features", fontsize=12)
        plt.ylabel("Loadings", fontsize=12)
        plt.title("PCA Loadings", fontsize=14)
        plt.xticks(np.arange(n_features), self.X.columns, rotation=90)
        plt.legend(loc="best")
        plt.tight_layout()

        if save:
            plt.savefig(
                os.path.join(self.plot_output_path, "pca_loadings.png"),
                format="png",
                bbox_inches="tight",
            )
        if show:
            plt.show()
        plt.close()

        return

    def _generate_greedy_ensemble_results_html(self, greedy_ensemble):
        """
        Generates an HTML table with the results of the greedy ensemble.

        Parameters
        ----------
        greedy_ensemble : sklearn.pipeline.Pipeline
            The greedy ensemble pipeline containing preprocessing steps and stacking classifier.

        Returns
        -------
        str
            An HTML string representing the results of the greedy ensemble.
        """
        if not greedy_ensemble:
            return ""
        greedy_ensemble = greedy_ensemble.named_steps["model"]
        results = self._score_model_with_metrics(greedy_ensemble)

        results_df = pd.DataFrame(
            [
                {
                    "model": "Greedy Ensemble",
                    **results,
                }
            ]
        )
        results_df = results_df.rename(
            columns={
                "model": "Model",
                "accuracy_score": "Accuracy",
                "balanced_accuracy_score": "Balanced Accuracy",
                "precision_score": "Precision",
                "recall_score": "Recall",
                "f1_score": "F1 Score",
                "jaccard_score": "Jaccard Score",
                "roc_auc_score": "ROC AUC",
            }
        )

        html_table = results_df.to_html(index=False)
        return html_table

    def evaluate_to_html(
        self,
        training_summary: pd.DataFrame,
    ):
        # Check if the training_summary is a DataFrame and not empty!:
        if (
            training_summary is None
            or not isinstance(training_summary, pd.DataFrame)  # noqa
            or training_summary.empty  # noqa
        ):
            raise ValueError(
                "Can't produce a HTML report because training_summary should be a DataFrame and not empty."
            )

        training_summary = training_summary.rename(
            columns={
                "model": "Model",
                "accuracy_score": "Accuracy",
                "balanced_accuracy_score": "Balanced Accuracy",
                "precision_score": "Precision",
                "recall_score": "Recall",
                "f1_score": "F1 Score",
                "jaccard_score": "Jaccard Score",
                "roc_auc_score": "ROC AUC",
                "duration": "Training Time [s]",
            }
        )
        # Sort the training_summary DataFrame by the score_metric column
        training_summary = training_summary.sort_values(
            by=training_summary.columns[1], ascending=False
        ).reset_index(drop=True)

        self.training_summary = training_summary

        styled_training_summary = training_summary.style.apply(
            _highlight_first_cell, axis=1
        )

        # Transform summary to HTML:
        training_summary_html = styled_training_summary.to_html()
        image_header_path = os.path.join(self.report_template_path, "mamut_header.png")
        base64_image = _get_base64_image(image_header_path)

        dataset_basic_list, feature_summary, class_distribution = (
            _generate_dataset_overview(self.X, self.y)
        )

        if self.binary:
            self._plot_roc_auc_curve(training_summary)
        else:
            self._plot_roc_auc_curve_multiclass(training_summary)

        self._plot_confusion_matrices(training_summary)
        self._plot_hyperparameter_tuning_history(training_summary)
        self._plot_feature_importances()
        best_model_name = training_summary.iloc[0]["Model"]
        best_model = self.models[best_model_name]

        self._plot_shap_beeswarm(best_model)

        if self.pca:
            self._plot_pca_loadings()

        # Load the Jinja2 template placed in report_template_path:
        env = Environment(loader=FileSystemLoader(self.report_template_path))
        template = env.get_template("report_template.html")

        # Render the template with the training_summary and save the HTML file
        time_signature = str(time.strftime(" %d %B %Y, %I:%M %p", time.localtime()))

        html_content = template.render(
            time_signature=time_signature,
            training_summary=training_summary_html,
            image_header=base64_image,
            experiment_setup=_generate_experiment_setup_table(),
            models_evaluated=_generate_models_list(self.excluded_models),
            optimizer=(
                "Tree-structured Parzen Estimator"
                if self.optimizer == "bayes"
                else "Random Search"
            ),
            metric=self.metric,
            n_trials=self.n_trials,
            best_model=training_summary.iloc[0]["Model"],
            basic_dataset_info=dataset_basic_list,
            feature_summary=feature_summary.to_html(index=False),
            class_distribution=class_distribution.to_html(index=False),
            feature_importance_method="Extra Trees Importances",
            pca=self.pca,
            binary=self.binary,
            is_ensemble=self.is_ensemble,
            ensemble_method="Stacking",
            ensemble_list=_generate_ensemble_list(self.greedy_ensemble),
            ensemble_summary=self._generate_greedy_ensemble_results_html(
                self.greedy_ensemble
            ),
            preprocessing_list=_generate_preprocessing_steps_html(
                self.preprocessing_steps
            ),
        )

        time_signature = datetime.strptime(
            time_signature.strip(), "%d %B %Y, %I:%M %p"
        ).strftime("%d-%m-%Y_%H-%M")
        with open(
            os.path.join(self.report_output_path, f"report_{time_signature}.html"), "w"
        ) as f:
            f.write(html_content)

        return html_content

    def _score_model_with_metrics(self, fitted_model):
        if not hasattr(fitted_model, "predict"):
            raise ValueError(
                "The model is not fitted and can not be scored with any metric."
            )

        y_pred = fitted_model.predict(self.X_test)
        y_pred_proba = fitted_model.predict_proba(self.X_test)
        if self.binary:
            y_pred_proba = y_pred_proba[:, 1]

        results = {
            "accuracy_score": accuracy_score(self.y_test, y_pred),
            "balanced_accuracy_score": balanced_accuracy_score(self.y_test, y_pred),
            "precision_score": precision_score(self.y_test, y_pred, average="weighted"),
            "recall_score": recall_score(self.y_test, y_pred, average="weighted"),
            "f1_score": f1_score(self.y_test, y_pred, average="weighted"),
            "jaccard_score": jaccard_score(self.y_test, y_pred, average="weighted"),
            "roc_auc_score": roc_auc_score(
                self.y_test, y_pred_proba, multi_class="ovr", average="weighted"
            ),
        }

        results = {
            self.metric: results.pop(self.metric),
            **results,
        }
        return results


def _highlight_first_cell(s):
    return [
        (
            "background-color: yellow"
            if (i == 0 and s.name == 0) or (i == 1 and s.name == 0)
            else ""
        )
        for i in range(len(s))
    ]



import ast
import time
import warnings
from copy import copy
from typing import Callable, List, Literal, Optional

import numpy as np
import optuna
import pandas as pd
from optuna.samplers import RandomSampler, TPESampler
from sklearn.ensemble import (  # noqa
    RandomForestClassifier,
    StackingClassifier,
    VotingClassifier,
)
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import LogisticRegression  # noqa
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    f1_score,
    jaccard_score,
    precision_score,
    recall_score,
    roc_auc_score,
)
from sklearn.model_selection import StratifiedKFold
from sklearn.naive_bayes import GaussianNB  # noqa
from sklearn.neighbors import KNeighborsClassifier  # noqa
from sklearn.neural_network import MLPClassifier  # noqa
from sklearn.svm import SVC  # noqa
from xgboost import XGBClassifier  # noqa

from mamut.utils.utils import adjust_search_spaces, model_param_dict, sample_parameter

optuna.logging.set_verbosity(optuna.logging.WARNING)
warnings.filterwarnings("ignore", category=ConvergenceWarning)


class ModelSelector:
    def __init__(
        self,
        X_train,
        y_train,
        X_test,
        y_test,
        score_metric: Callable,
        exclude_models: Optional[List[str]] = None,
        optimization_method: Literal["random_search", "bayes"] = "bayes",
        n_iterations: int = 50,
        random_state: Optional[int] = 42,
    ):

        self.X_train = X_train
        self.y_train = y_train
        self.X_test = X_test
        self.y_test = y_test
        if not exclude_models:
            exclude_models = []

        self.models = [
            (
                eval(model)(random_state=random_state)
                if "random_state" in eval(model)().get_params()
                else eval(model)()
            )
            for model in model_param_dict.keys()
            if model not in exclude_models
        ]
        self.n_classes_ = len(np.unique(y_train))
        self.binary = True if self.n_classes_ == 2 else False
        self.score_metric_name = copy(score_metric.__name__)
        self.roc = self.score_metric_name == "roc_auc_score"
        if self.roc:
            self.score_metric = lambda y_true, y_pred: score_metric(
                y_true, y_pred, multi_class="ovr", average="weighted"
            )
        else:
            self.score_metric = lambda y_true, y_pred: score_metric(
                y_true.reshape(-1, 1), y_pred.reshape(-1, 1), average="weighted"
            )

        self.optuna_sampler = (
            TPESampler(seed=random_state)
            if optimization_method == "bayes"
            else RandomSampler(seed=random_state)
        )
        self.n_iterations = n_iterations
        self.SKF_ = StratifiedKFold(n_splits=5, shuffle=True, random_state=random_state)

    def objective(self, trial, model):
        if model.__class__.__name__ in model_param_dict:
            param_grid = model_param_dict[model.__class__.__name__]
        else:
            raise ValueError(f"Model {model.__class__.__name__} not supported")

        param = {
            param_name: sample_parameter(trial, param_name, value)
            for param_name, value in param_grid.items()
        }

        param = adjust_search_spaces(param, model)

        model.set_params(**param)

        cv_scores = []
        for train_idx, val_idx in self.SKF_.split(self.X_train, self.y_train):
            X_train_fold, X_val_fold = self.X_train[train_idx], self.X_train[val_idx]
            y_train_fold, y_val_fold = self.y_train[train_idx], self.y_train[val_idx]

            model = model.__class__(**model.get_params())
            model.fit(X_train_fold, y_train_fold)
            val_pred = (
                model.predict_proba(X_val_fold)
                if self.roc
                else model.predict(X_val_fold)
            )
            if self.binary and self.roc:
                val_pred = val_pred[:, 1]

            cv_scores.append(self.score_metric(y_val_fold, val_pred))
        mean_cv_score = np.mean(cv_scores)

        return mean_cv_score

    def optimize_model(self, model):
        study = optuna.create_study(direction="maximize", sampler=self.optuna_sampler)
        start_time = time.time()
        study.optimize(
            lambda trial: self.objective(trial, model),
            n_trials=self.n_iterations,
            show_progress_bar=True,
        )
        end_time = time.time()
        duration = end_time - start_time
        best_params = study.best_params
        hidden_sizes = best_params.get("hidden_layer_sizes")
        if isinstance(hidden_sizes, str):
            try:
                best_params["hidden_layer_sizes"] = ast.literal_eval(hidden_sizes)
            except (ValueError, SyntaxError):
                pass
        return best_params, study.best_value, duration, study

    def compare_models(self):
        best_model = None
        score_for_best_model = 0
        params_for_best_model = None
        fitted_models = {}
        training_summary = pd.DataFrame()
        studies = {}

        for model in self.models:
            print(f"Optimizing model: {model.__class__.__name__}")
            params, score, duration, study = self.optimize_model(model)
            print(
                f"Best parameters: {params}, score: {score:.4f} {self.score_metric_name}\n"
            )

            model.set_params(**params)
            model = model.__class__(**model.get_params())
            model.fit(self.X_train, self.y_train)
            fitted_models[model.__class__.__name__] = model
            studies[model.__class__.__name__] = study

            if self.roc:
                if self.binary:
                    score_on_test = self.score_metric(
                        self.y_test.values, model.predict_proba(self.X_test)[:, 1]
                    )
                else:
                    score_on_test = self.score_metric(
                        self.y_test.values, model.predict_proba(self.X_test)
                    )
            else:
                score_on_test = self.score_metric(
                    self.y_test.values, model.predict(self.X_test)
                )

            if score_on_test > score_for_best_model:
                score_for_best_model = score_on_test
                best_model = model
                params_for_best_model = params

            scores_on_test = self._score_model_with_metrics(model)

            training_summary = pd.concat(
                [
                    training_summary,
                    pd.DataFrame(
                        [
                            {
                                "model": model.__class__.__name__,
                                **scores_on_test,
                                "duration": duration,
                            }
                        ]
                    ),
                ],
                ignore_index=True,
            )

        print(
            f"Found best model: {best_model.__class__.__name__} with parameters {params_for_best_model} \n"
            f"and score {score_for_best_model:.4f} {self.score_metric_name}. \n"
            f"To access your best model use: mamut.best_model_ field. \n"
            f"To create a powerful ensemble of models use: create_greedy_ensemble() function. \n"
        )

        return (
            best_model,
            params_for_best_model,
            score_for_best_model,
            fitted_models,
            training_summary,
            studies,
        )

    def _score_model_with_metrics(self, fitted_model):
        if not hasattr(fitted_model, "predict"):
            raise ValueError(
                "The model is not fitted and can not be scored with any metric."
            )

        y_pred = fitted_model.predict(self.X_test)
        y_pred_proba = fitted_model.predict_proba(self.X_test)
        if self.binary:
            y_pred_proba = y_pred_proba[:, 1]

        results = {
            "accuracy_score": accuracy_score(self.y_test, y_pred),
            "balanced_accuracy_score": balanced_accuracy_score(self.y_test, y_pred),
            "precision_score": precision_score(
                self.y_test, y_pred, average="weighted", zero_division=0
            ),
            "recall_score": recall_score(
                self.y_test, y_pred, average="weighted", zero_division=0
            ),
            "f1_score": f1_score(
                self.y_test, y_pred, average="weighted", zero_division=0
            ),
            "jaccard_score": jaccard_score(
                self.y_test, y_pred, average="weighted", zero_division=0
            ),
            "roc_auc_score": roc_auc_score(
                self.y_test, y_pred_proba, multi_class="ovr", average="weighted"
            ),
        }

        results = {
            self.score_metric_name: results.pop(self.score_metric_name),
            **results,
        }
        return results



import warnings
from typing import List, Literal, Optional

import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline

from mamut.preprocessing.handlers import (
    handle_categorical,
    handle_extraction,
    handle_imbalanced,
    handle_missing_categorical,
    handle_missing_numeric,
    handle_outliers,
    handle_scaling,
    handle_selection,
    handle_skewed,
)

warnings.filterwarnings(
    "ignore",
    category=DeprecationWarning,
    module="numpy\\.ma\\.extras",
)
warnings.filterwarnings(
    "ignore",
    category=DeprecationWarning,
    message="Conversion of an array with ndim > 0 to a scalar is deprecated.*",
)


class Preprocessor:
    """
    A class used to preprocess data for machine learning models.

    Attributes
    ----------
    numeric_features : Optional[List[str]]
        List of numeric feature names.
    categorical_features : Optional[List[str]]
        List of categorical feature names.
    num_imputation : Literal["iterative", "knn", "mean", "median", "constant"]
        Method for numeric imputation.
    cat_imputation : Literal["most_frequent", "constant"]
        Method for categorical imputation.
    scaling : Literal["standard", "robust"]
        Method for scaling numeric features.
    feature_selection : bool
        Whether to perform feature selection.
    pca : bool
        Whether to perform PCA for feature extraction.
    imbalanced_resampling : bool
        Whether to perform resampling to handle imbalanced data.
    resampling_strategy : Literal["SMOTE", "undersample", "combine"]
        Strategy for resampling imbalanced data.
    skew_threshold : float
        Threshold for skewness correction.
    pca_threshold : float
        Threshold for PCA feature extraction.
    selection_threshold : float
        Threshold for feature selection.
    imbalance_threshold : float
        Threshold for detecting imbalanced data.
    random_state : Optional[int]
        Random state for reproducibility.

    Methods
    -------
    fit_transform(X: pd.DataFrame, y: pd.Series) -> (np.ndarray, np.ndarray, Pipeline)
        Fits the preprocessor and transforms the data.
    transform(X: pd.DataFrame) -> np.ndarray
        Transforms the data using the fitted preprocessor.
    report() -> dict
        Returns a report of the preprocessing steps.
    _check_fitted()
        Checks if the preprocessor has been fitted.
    """

    def __init__(
        self,
        numeric_features: Optional[List[str]] = None,
        categorical_features: Optional[List[str]] = None,
        num_imputation: Literal[
            "iterative", "knn", "mean", "median", "constant"
        ] = "knn",
        cat_imputation: Literal["most_frequent", "constant"] = "most_frequent",
        scaling: Literal["standard", "robust"] = "standard",
        feature_selection: bool = False,
        pca: bool = False,
        imbalanced_resampling: bool = True,
        resampling_strategy: Literal["SMOTE", "undersample", "combine"] = "SMOTE",
        skew_threshold: float = 1,
        pca_threshold: float = 0.95,
        selection_threshold: float = 0.05,
        imbalance_threshold: float = 0.10,
        random_state: Optional[int] = 42,
    ) -> None:
        """
        Constructs all the necessary attributes for the Preprocessor object.

        Parameters
        ----------
        numeric_features : Optional[List[str]]
            List of numeric feature names.
        categorical_features : Optional[List[str]]
            List of categorical feature names.
        num_imputation : Literal["iterative", "knn", "mean", "median", "constant"]
            Method for numeric imputation.
        cat_imputation : Literal["most_frequent", "constant"]
            Method for categorical imputation.
        scaling : Literal["standard", "robust"]
            Method for scaling numeric features.
        feature_selection : bool
            Whether to perform feature selection.
        pca : bool
            Whether to perform PCA for feature extraction.
        imbalanced_resampling : bool
            Whether to perform resampling to handle imbalanced data.
        resampling_strategy : Literal["SMOTE", "undersample", "combine"]
            Strategy for resampling imbalanced data.
        skew_threshold : float
            Threshold for skewness correction.
        pca_threshold : float
            Threshold for PCA feature extraction.
        selection_threshold : float
            Threshold for feature selection.
        imbalance_threshold : float
            Threshold for detecting imbalanced data.
        random_state : Optional[int]
            Random state for reproducibility.
        """
        self.numeric_features = numeric_features
        self.categorical_features = categorical_features
        self.num_imputation = num_imputation
        self.cat_imputation = cat_imputation
        self.feature_selection = feature_selection
        self.pca = pca
        self.random_state = random_state
        self.scaling = scaling
        self.pca_threshold = pca_threshold
        self.selection_threshold = selection_threshold
        self.imbalance_threshold = imbalance_threshold
        self.imbalanced_resampling = imbalanced_resampling
        self.resampling_strategy = resampling_strategy
        self.skew_threshold = skew_threshold

        self.imbalanced_ = None
        self.missing_ = None
        self.imbalanced_trans_ = None
        self.outlier_trans_ = None
        self.missing_num_trans_ = None
        self.missing_cat_trans_ = None
        self.cat_trans_ = None
        self.skew_trans_ = None
        self.skewed_ = None
        self.scaler_ = None
        self.sel_trans_ = None
        self.ext_trans_ = None
        self.fitted = False
        self.skewed_feature_names_ = None
        self.selected_features_ = None
        self.pca_loadings_ = None
        self.missing_numeric_ = None
        self.missing_categorical_ = None
        self.has_numeric_ = None
        self.has_categorical_ = None
        self.ohe_feature_names_ = None
        self.report_ = None
        self.n_missing_numeric = None
        self.n_missing_categorical = None
        self.lambdas_ = None
        self.feature_importances_ = None

    def fit_transform(
        self, X: pd.DataFrame, y: pd.Series
    ) -> (np.ndarray, np.ndarray, Pipeline):
        """
        Fits the preprocessor and transforms the data.

        Parameters
        ----------
        X : pd.DataFrame
            The input features.
        y : pd.Series
            The target variable.

        Returns
        -------
        np.ndarray
            Transformed features.
        np.ndarray
            Transformed target variable.
        Pipeline
            The fitted pipeline.
        """
        self.report_ = dict()

        if not self.numeric_features:
            self.numeric_features = X.select_dtypes(include="number").columns.tolist()
        if not self.categorical_features:
            self.categorical_features = X.select_dtypes(
                exclude="number"
            ).columns.tolist()

        self.report_["features"] = {
            "numeric": self.numeric_features,
            "categorical": self.categorical_features,
        }

        if y.value_counts(normalize=True).min() < self.imbalance_threshold:
            self.imbalanced_ = True
        X, y = X.copy(), y.copy()

        self.has_numeric_ = len(self.numeric_features) > 0
        self.has_categorical_ = len(self.categorical_features) > 0

        if self.has_numeric_:
            self.n_missing_numeric = X[self.numeric_features].isnull().sum().sum()
            if self.n_missing_numeric > 0:
                self.missing_numeric_ = True
                X, self.missing_num_trans_ = handle_missing_numeric(
                    X, self.numeric_features, self.num_imputation
                )

        if self.has_categorical_:
            self.n_missing_categorical = (
                X[self.categorical_features].isnull().sum().sum()
            )
            if self.n_missing_categorical > 0:
                self.missing_categorical_ = True
                X, self.missing_cat_trans_ = handle_missing_categorical(
                    X, self.categorical_features, self.cat_imputation
                )

        self.missing_ = self.missing_numeric_ or self.missing_categorical_

        if self.missing_:
            self.report_["imputation"] = {}
            if self.missing_numeric_:
                self.report_["imputation"]["numeric"] = {
                    "transformer": self.missing_num_trans_.__class__.__name__,
                    "n_missing_numeric": self.n_missing_numeric,
                }
            if self.missing_categorical_:
                self.report_["imputation"]["categorical"] = {
                    "transformer": self.missing_cat_trans_.__class__.__name__,
                    "n_missing_categorical": self.n_missing_categorical,
                }

        if self.has_numeric_:
            n_row_before = X.shape[0]
            X, y, self.outlier_trans_ = handle_outliers(
                X, y, self.numeric_features, random_state=self.random_state
            )
            n_row_after = X.shape[0]
            self.report_["removing_outliers"] = {
                "transformer": self.outlier_trans_.__class__.__name__,
                "n_outliers_removed": n_row_before - n_row_after,
            }

        if self.has_categorical_:
            X, self.cat_trans_, self.ohe_feature_names_ = handle_categorical(
                X, self.categorical_features
            )
            self.report_["category_encoding"] = {
                "transformer": self.cat_trans_.__class__.__name__,
                "encoded_feature_names": self.ohe_feature_names_,
            }

        if self.has_numeric_:
            (
                X,
                self.skew_trans_,
                self.skewed_feature_names_,
                self.lambdas_,
            ) = handle_skewed(X, self.numeric_features, threshold=self.skew_threshold)
            self.report_["skew_transform"] = {
                "transformer": self.skew_trans_.__class__.__name__,
                "method": self.skew_trans_.method,
                "skewed_feature_names": self.skewed_feature_names_,
                "lambdas": self.lambdas_,
            }
        else:
            self.skewed_feature_names_ = []
            self.lambdas_ = []

        if self.has_numeric_:
            X, self.scaler_ = handle_scaling(X, self.numeric_features, self.scaling)
            self.report_["scaling"] = {
                "transformer": self.scaler_.__class__.__name__,
            }

        if self.feature_selection:
            X, self.sel_trans_, self.selected_features_, self.feature_importances_ = (
                handle_selection(
                    X,
                    y,
                    threshold=self.selection_threshold,
                    random_state=self.random_state,
                )
            )
            self.report_["feature_selection"] = {
                "transformer": self.sel_trans_.__class__.__name__,
                "estimator": self.sel_trans_.estimator_.__class__.__name__,
                "selected_features": self.selected_features_,
                "feature_importances": self.feature_importances_,
            }

        if self.pca:
            n_features_before = X.shape[1]
            X, self.ext_trans_, self.pca_loadings_ = handle_extraction(
                X, threshold=self.pca_threshold, random_state=self.random_state
            )
            n_features_after = X.shape[1]
            self.report_["feature_extraction"] = {
                "transformer": self.ext_trans_.__class__.__name__,
                "pca_loadings": self.pca_loadings_,
                "n_features_before": n_features_before,
                "n_features_after": n_features_after,
            }

        if self.imbalanced_resampling and self.imbalanced_:
            n_row_before = X.shape[0]
            X, y, self.imbalanced_trans_ = handle_imbalanced(
                X, y, self.resampling_strategy, random_state=self.random_state
            )
            if self.imbalanced_trans_ is not None:
                n_row_after = X.shape[0]
                self.report_["imbalanced_resampling"] = {
                    "transformer": self.imbalanced_trans_.__class__.__name__,
                    "n_resampled": n_row_after - n_row_before,
                }
            else:
                self.report_["imbalanced_resampling"] = {
                    "skipped": True,
                    "reason": "Insufficient minority samples for SMOTE.",
                    "strategy": self.resampling_strategy,
                }

        self.skewed_ = len(self.skewed_feature_names_) > 0
        self.fitted = True

        if isinstance(X, pd.DataFrame):
            X = X.values

        if isinstance(y, pd.Series):
            y = y.values

        return X, y

    def transform(self, X: pd.DataFrame) -> np.ndarray:
        """
        Transforms the data using the fitted preprocessor.

        Parameters
        ----------
        X : pd.DataFrame
            The input features.

        Returns
        -------
        np.ndarray
            Transformed features.
        """
        self._check_fitted()
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input data must be a pandas DataFrame.")
        X = X.copy()
        if self.missing_numeric_:
            X[self.numeric_features] = self.missing_num_trans_.transform(
                X[self.numeric_features]
            )

        if self.missing_categorical_:
            X[self.categorical_features] = self.missing_cat_trans_.transform(
                X[self.categorical_features]
            )

        if self.has_categorical_:
            encoded_features = self.cat_trans_.transform(X[self.categorical_features])
            encoded_features_df = pd.DataFrame(
                encoded_features,
                columns=self.ohe_feature_names_,
                index=X.index,
            )
            X = X.drop(columns=self.categorical_features).join(encoded_features_df)

        if self.skewed_:
            X[self.skewed_feature_names_] = self.skew_trans_.transform(
                X[self.skewed_feature_names_]
            )

        if self.has_numeric_:
            X[self.numeric_features] = self.scaler_.transform(X[self.numeric_features])

        if self.feature_selection:
            X = self.sel_trans_.transform(X)
            X = pd.DataFrame(X, columns=self.selected_features_)

        if self.pca:
            X = self.ext_trans_.transform(X)

        if isinstance(X, pd.DataFrame):
            X = X.values

        return X

    def report(self):
        """
        Returns a report of the preprocessing steps.

        Returns
        -------
        dict
            A dictionary containing the report of the preprocessing steps.
        """
        self._check_fitted()
        return self.report_

    def _check_fitted(self):
        """
        Checks if the preprocessor has been fitted.

        Raises
        ------
        RuntimeError
            If the preprocessor has not been fitted.
        """
        if not self.fitted:
            raise RuntimeError("Preprocessor has not been fitted.")



from sklearn.linear_model import LogisticRegression
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    f1_score,
    jaccard_score,
    precision_score,
    recall_score,
    roc_auc_score,
)

lr_params = {
    "C": (1e-4, 1e4, "log"),
    "l1_ratio": (1e-4, 1.0, "log"),
    "class_weight": (["balanced"], "categorical"),
    "max_iter": (1000, 1000, "int"),
    "solver": (["saga", "lbfgs", "liblinear"], "categorical"),
}

tree_params = {
    "n_estimators": (10, 1000, "int"),
    "criterion": (["gini", "entropy", "log_loss"], "categorical"),
    "bootstrap": ([True], "categorical"),
    "max_samples": (0.5, 1, "float"),
    "max_features": (0.1, 0.9, "float"),
    "min_samples_leaf": (0.05, 0.25, "float"),
}

xgb_params = {
    "n_estimators": (10, 1000, "int"),
    "learning_rate": (1e-4, 0.4, "log"),
    "subsample": (0.25, 1.0, "float"),
    "booster": (["gbtree"], "categorical"),
    "max_depth": (1, 15, "int"),
    "min_child_weight": (1, 128, "float"),
    "colsample_bytree": (0.2, 1.0, "float"),
    "colsample_bylevel": (0.2, 1.0, "float"),
    "reg_alpha": (1e-4, 512.0, "log"),
    "reg_lambda": (1e-3, 1e3, "log"),
}

svc_params = {
    "C": (1e-4, 1e4, "log"),
    "kernel": (["linear", "poly", "rbf", "sigmoid"], "categorical"),
    "gamma": (1e-4, 1.0, "log"),
    "class_weight": (["balanced"], "categorical"),
    "probability": ([True], "categorical"),
}

mlp_params = {
    "hidden_layer_sizes": (
        [(32,), (64,), (128,), (256,), (32, 16), (32, 32), (64, 32), (64, 64)],
        "categorical",
    ),
    "activation": (["identity", "logistic", "tanh", "relu"], "categorical"),
    "solver": (["lbfgs", "sgd", "adam"], "categorical"),
    "alpha": (1e-5, 1e-2, "log"),
    "learning_rate": (["constant", "invscaling", "adaptive"], "categorical"),
    "learning_rate_init": (1e-4, 1e-1, "log"),
    "power_t": (0.1, 0.9, "float"),
    "max_iter": (100, 200, "int"),
    "momentum": (0.5, 0.9, "float"),
}

gnb_params = {
    "var_smoothing": (1e-9, 1e-5, "log"),
}

knn_params = {
    "n_neighbors": (1, 30, "int"),
}

model_param_dict = {
    "LogisticRegression": lr_params,
    "RandomForestClassifier": tree_params,
    "SVC": svc_params,
    "XGBClassifier": xgb_params,
    "MLPClassifier": mlp_params,
    "GaussianNB": gnb_params,
    "KNeighborsClassifier": knn_params,
}

metric_dict = {
    "accuracy": accuracy_score,
    "precision": precision_score,
    "recall": recall_score,
    "f1": f1_score,
    "balanced_accuracy": balanced_accuracy_score,
    "jaccard": jaccard_score,
    "roc_auc_score": roc_auc_score,
}

preprocessing_steps = {
    "IsolationForest": (
        "Outlier detection",
        "Detects and removes outliers from the dataset.",
    ),
    "SimpleImputer": ("Imputation", "Fills missing values in the dataset."),
    "PowerTransformer": (
        "Transformation",
        "Applies power transformation to reduce skewness.",
    ),
    "OneHotEncoder": (
        "Encoding",
        "Encodes categorical features using one-hot encoding.",
    ),
    "SelectFromModel": (
        "Feature selection",
        "Selects features based on importance weights.",
    ),
    "PCA": (
        "Dimensionality reduction",
        "Reduces dimensionality of the dataset using Principal Component Analysis.",
    ),
    "ExtraTreesClassifier": (
        "Feature selection",
        "Classifier used for feature selection.",
    ),
    "StandardScaler": (
        "Scaling",
        "Standardizes features by removing the mean and scaling to unit variance.",
    ),
    "RobustScaler": (
        "Scaling",
        "Scales features using statistics that are robust to outliers.",
    ),
    "KNNImputer": ("Imputation", "Fills missing values using k-Nearest Neighbors."),
    "IterativeImputer": (
        "Imputation",
        "Fills missing values using iterative imputation.",
    ),
}


def sample_parameter(trial, param_name, value):
    """Sample a parameter value based on its distribution type."""
    if len(value) == 3:
        low, high, dist_type = value
        if dist_type == "log":
            return trial.suggest_float(param_name, low, high, log=True)
        elif dist_type == "float":
            return trial.suggest_float(param_name, low, high)
        else:
            return trial.suggest_int(param_name, low, high)
    elif len(value) == 2:
        options, dist_type = value
        if any(isinstance(option, tuple) for option in options):
            option_map = {repr(option): option for option in options}
            choice = trial.suggest_categorical(param_name, list(option_map.keys()))
            return option_map[choice]
        return trial.suggest_categorical(param_name, options)
    else:
        raise ValueError("Invalid hyperparameter search space.")


def adjust_search_spaces(param_dict, model):
    if isinstance(model, LogisticRegression):
        if param_dict["solver"] == "saga":
            param_dict["penalty"] = "elasticnet"
        else:
            param_dict["penalty"] = "l2"
            param_dict["l1_ratio"] = None

    return param_dict



from typing import List, Literal

import numpy as np
import pandas as pd
from imblearn.combine import SMOTETomek
from imblearn.over_sampling import SMOTE
from scipy.stats import skew
from sklearn.decomposition import PCA
from sklearn.ensemble import ExtraTreesClassifier, IsolationForest
from sklearn.feature_selection import SelectFromModel
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import OneHotEncoder, PowerTransformer

import mamut.preprocessing.settings as settings


def handle_outliers(
    X: pd.DataFrame,
    y: pd.Series,
    feature_names: List[str],
    contamination: float = 0.01,
    random_state: int = 42,
) -> (pd.DataFrame, pd.Series, IsolationForest):
    """
    Handles outliers in the dataset using IsolationForest.

    Parameters:
        X: pd.DataFrame
            Feature matrix.
        y: pd.Series
            Target array.
        feature_names: List[str]
            Names of the features in the dataset.
        contamination: float
            The proportion of outliers in the data.
        random_state: int
            Seed for reproducibility.

    Returns:
        X_filtered: pd.DataFrame
            Feature matrix with outliers removed.
        y_filtered: pd.Series
            Target array with outliers removed.
        transformer: IsolationForest
            Fitted IsolationForest model.
    """
    X, y = X.copy(), y.copy()
    iso_forest = IsolationForest(contamination=contamination, random_state=random_state)
    outliers = iso_forest.fit_predict(X[feature_names])
    mask = outliers == 1

    return X[mask], y[mask], iso_forest


def handle_imbalanced(
    X: pd.DataFrame,
    y: pd.Series,
    strategy: Literal["SMOTE", "undersample", "combine"],
    random_state=42,
) -> (pd.DataFrame, pd.Series, object):
    """
    Balances an imbalanced dataset using techniques from imbalanced-learn.

    Parameters:
        X: pd.DataFrame
            Feature matrix.
        y: pd.Series
            Target array.
        strategy: Literal["SMOTE", "undersample", "combine"]
            Resampling method to use. Options:
            - 'SMOTE': Synthetic Minority Oversampling Technique.
            - 'undersample': Random undersampling of majority class.
            - 'combine': SMOTE with Tomek links.
        random_state: int
            Seed for reproducibility.

    Returns:
        X_resampled: pd.DataFrame
            Feature matrix after resampling.
        y_resampled: pd.Series
            Target array after resampling.
        transformer: object
            Fitted resampling method instance.
    """
    if strategy not in settings.resampler_mapping.keys():
        raise ValueError(
            f"Invalid resampling strategy, choose from {settings.resampler_mapping.keys()}."
        )

    y_series = y if isinstance(y, pd.Series) else pd.Series(y)
    minority_count = y_series.value_counts().min()
    if strategy in {"SMOTE", "combine"}:
        k_neighbors = min(5, minority_count - 1)
        if k_neighbors < 1:
            return X, y, None
        smote = SMOTE(random_state=random_state, k_neighbors=k_neighbors)
        if strategy == "SMOTE":
            resampler = smote
        else:
            resampler = SMOTETomek(random_state=random_state, smote=smote)
    else:
        resampler = settings.resampler_mapping[strategy](random_state=random_state)
    X_resampled, y_resampled = resampler.fit_resample(X, y)

    return X_resampled, y_resampled, resampler


def handle_skewed(
    X: pd.DataFrame, feature_names: List[str], threshold: float = 1
) -> (pd.DataFrame, PowerTransformer, List[str]):
    """
    Handles skewed features in the dataset using PowerTransformer.

    Parameters:
        X: pd.DataFrame
            Feature matrix.
        feature_names: List[str]
            Names of the features in the dataset.
        threshold: float
            Threshold for skewness.

    Returns:
        X_transformed: pd.DataFrame
            Feature matrix with skewed features transformed.
        transformer: PowerTransformer
            Fitted PowerTransformer model.
        skewed_feature_names: List[str]
            Names of the skewed features that were transformed.
        lambdas: List[float]
            Lambda values for the transformed features.
    """
    X = X.copy()
    skewed_feature_names = []
    pt = PowerTransformer(method="yeo-johnson", standardize=False)
    for feature in feature_names:
        feature_skewness = skew(X[feature])
        if abs(feature_skewness) > threshold:
            skewed_feature_names.append(feature)

    lambdas = []
    if len(skewed_feature_names) > 0:
        X[skewed_feature_names] = pt.fit_transform(X[skewed_feature_names])
        lambdas = pt.lambdas_

    return X, pt, skewed_feature_names, lambdas


def handle_missing_numeric(
    X: pd.DataFrame,
    feature_names: List[str],
    strategy: Literal["iterative", "knn", "mean", "median", "constant"],
) -> (pd.DataFrame, object):
    """
    Handles missing numeric values in the dataset using specified imputation strategy.

    Parameters:
        X: pd.DataFrame
            Feature matrix.
        feature_names: List[str]
            Names of the numeric features in the dataset.
        strategy: Literal["iterative", "knn", "mean", "median", "constant"]
            Imputation strategy to use.

    Returns:
        X_imputed: pd.DataFrame
            Feature matrix with missing numeric values imputed.
        imputer: object
            Fitted imputer model.
    """
    if strategy not in settings.imputer_mapping.keys():
        raise ValueError(
            f"Invalid imputation strategy, choose from {settings.imputer_mapping.keys()}."
        )

    X = X.copy()
    imputer = settings.imputer_mapping[strategy]()
    imputer.fit(X[feature_names])
    X[feature_names] = imputer.transform(X[feature_names])

    return X, imputer


def handle_missing_categorical(
    X: pd.DataFrame,
    feature_names: List[str],
    strategy: Literal["most_frequent", "constant"],
) -> (pd.DataFrame, SimpleImputer):
    """
    Handles missing categorical values in the dataset using specified imputation strategy.

    Parameters:
        X: pd.DataFrame
            Feature matrix.
        feature_names: List[str]
            Names of the categorical features in the dataset.
        strategy: Literal["most_frequent", "constant"]
            Imputation strategy to use.

    Returns:
        X_imputed: pd.DataFrame
            Feature matrix with missing categorical values imputed.
        imputer: SimpleImputer
            Fitted SimpleImputer model.
    """
    X = X.copy()
    imputer = SimpleImputer(strategy=strategy)
    imputer.fit(X[feature_names])
    X[feature_names] = imputer.transform(X[feature_names])

    return X, imputer


def handle_categorical(
    X: pd.DataFrame, feature_names: List[str]
) -> (pd.DataFrame, OneHotEncoder):
    """
    Handles categorical features in the dataset using OneHotEncoder.

    Parameters:
        X: pd.DataFrame
            Feature matrix.
        feature_names: List[str]
            Names of the categorical features in the dataset.

    Returns:
        X_encoded: pd.DataFrame
            Feature matrix with categorical features encoded.
        encoder: OneHotEncoder
            Fitted OneHotEncoder model.
        ohe_feature_names: List[str]
            Names of the one-hot encoded features.
    """
    X = X.copy()
    encoder = OneHotEncoder(drop="first", handle_unknown="error", sparse_output=False)
    encoder.fit(X[feature_names])
    encoded_features = encoder.transform(X[feature_names])
    ohe_feature_names = encoder.get_feature_names_out(feature_names)
    encoded_features_df = pd.DataFrame(
        encoded_features,
        columns=ohe_feature_names,
        index=X.index,
    )
    X = X.drop(columns=feature_names).join(encoded_features_df)

    return X, encoder, ohe_feature_names


def handle_scaling(
    X: pd.DataFrame, feature_names: List[str], strategy: Literal["standard", "robust"]
) -> (pd.DataFrame, object):
    """
    Handles scaling of features in the dataset using specified scaling strategy.

    Parameters:
        X: pd.DataFrame
            Feature matrix.
        feature_names: List[str]
            Names of the features to be scaled.
        strategy: Literal["standard", "robust"]
            Scaling strategy to use.

    Returns:
        X_scaled: pd.DataFrame
            Feature matrix with scaled features.
        scaler: object
            Fitted scaler instance.
    """
    if strategy not in ["standard", "robust"]:
        raise ValueError(
            f"Invalid scaling strategy, choose from {settings.scaler_mapping.keys()}."
        )

    X = X.copy()
    scaler = settings.scaler_mapping[strategy]()
    scaler.fit(X[feature_names])
    X[feature_names] = scaler.transform(X[feature_names])

    return X, scaler


def handle_selection(
    X: pd.DataFrame, y: pd.Series, threshold: float = 0.05, random_state: int = 42
) -> (pd.DataFrame, SelectFromModel, List[str]):
    """
    Handles feature selection using ExtraTreesClassifier.

    Parameters:
        X: pd.DataFrame
            Feature matrix.
        y: pd.Series
            Target array.
        threshold: float
            Threshold for feature selection.
        random_state: int
            Seed for reproducibility.

    Returns:
        X_selected: pd.DataFrame
            Feature matrix with selected features.
        selector: SelectFromModel
            Fitted SelectFromModel instance.
        selected_features: List[str]
            Names of the selected features.
        feature_importances: np.ndarray
            Feature importances from the ExtraTreesClassifier.
    """
    X = X.copy()
    selector = SelectFromModel(
        ExtraTreesClassifier(random_state=random_state), threshold=threshold
    )
    selector.fit(X, y)
    X_selected = selector.transform(X)
    selected_features = X.columns[selector.get_support()]
    X_selected_df = pd.DataFrame(X_selected, columns=selected_features, index=X.index)

    feature_importances = selector.estimator_.feature_importances_

    return X_selected_df, selector, selected_features, feature_importances


def handle_extraction(
    X: pd.DataFrame, threshold: float = 0.95, random_state: int = 42
) -> (np.ndarray, PCA):
    """
    Handles feature extraction using PCA.

    Parameters:
        X: pd.DataFrame
            Feature matrix.
        threshold: float
            Threshold for PCA.
        random_state: int
            Seed for reproducibility.

    Returns:
        X_extracted: np.ndarray
            Feature matrix after PCA transformation.
        extractor: PCA
            Fitted PCA instance.
        loadings: np.ndarray
            Loadings of the PCA components.
    """
    X = X.copy()
    extractor = PCA(
        n_components=threshold, svd_solver="full", random_state=random_state
    )
    extractor.fit(X)
    X_extracted = extractor.transform(X)
    loadings = extractor.components_

    return X_extracted, extractor, loadings
